# pylint: disable=pointless-statement,no-self-use,using-constant-test

import craftier


class SquareTransformer(craftier.CraftierTransformer):
    def square_before(self, x):
        x * x

    def square_after(self, x):
        x ** 2


class IfTrueTransformer(craftier.CraftierTransformer):
    def if_true_before(self, x, y):
        x if True else y

    def if_true_after(self, x):
        x



import contextlib
import dataclasses
import inspect
from typing import (
    Dict,
    Generator,
    List,
    Mapping,
    MutableMapping,
    Optional,
    Protocol,
    Sequence,
    Type,
    Union,
)

import libcst
import libcst.codemod
import libcst.metadata
from loguru import logger


class _LeaveMethod(Protocol):
    __self__: "ContextAwareTransformer"
    __name__: str
    __qualname__: str

    def __call__(
        self, original_node: libcst.CSTNodeT, updated_node: libcst.CSTNodeT
    ) -> Union[libcst.CSTNodeT, libcst.RemovalSentinel]:
        ...


CONTEXT_KEY = "craftier"


class ContextAwareTransformer(libcst.codemod.Codemod, libcst.CSTTransformer):
    """A lean replacement of libcst.codemod.ContextAwareTransformer.

    It just extends from `libcst.CSTTransformer` and not
    `libcst.matchers.MatcherDecoratableTransformer`, which reduces the checks
    performed when visiting nodes.
    """

    def __init__(self, context: libcst.codemod.CodemodContext) -> None:
        libcst.codemod.Codemod.__init__(self, context)
        libcst.CSTTransformer.__init__(self)

    def transform_module_impl(self, tree: libcst.Module) -> libcst.Module:
        return tree.visit(self)

    def get_leave_funcs(self) -> Mapping[str, _LeaveMethod]:
        """Return all the valid on_leave methods."""
        methods = inspect.getmembers(
            self,
            lambda m: (
                inspect.ismethod(m)
                and m.__name__.startswith("leave_")
                and not getattr(m, "_is_no_op", False)
            ),
        )

        return dict(methods)


class BatchedCodemod(libcst.codemod.Codemod, libcst.CSTTransformer):
    """Codemod which runs multiple transforms at the same time."""

    def __init__(
        self,
        context: libcst.codemod.CodemodContext,
        transformers: Sequence[Type[ContextAwareTransformer]],
        max_executions: int = 10,
    ):
        libcst.codemod.Codemod.__init__(self, context)
        libcst.CSTTransformer.__init__(self)
        self.max_executions = max_executions
        self.leave_methods: MutableMapping[str, List[_LeaveMethod]] = {}
        self.transformers = transformers
        self._batched_transformer: Optional[_BatchedTransformer] = None

    def transform_module_impl(self, tree: libcst.Module) -> libcst.Module:
        """Transform the tree.

        Note: we do not use should_allow_multiple_passes, as that approach
        compares the trees using an expensive deep compare operation. Instead,
        we use the information stored in the context by our transforms.

        This allow us to shave about 10% of the run time.
        """
        if not self._batched_transformer:
            leave_methods: Dict[str, List[_LeaveMethod]] = {}
            for transform_class in self.transformers:
                for name, method in (
                    transform_class(self.context).get_leave_funcs().items()
                ):
                    leave_methods.setdefault(
                        name.replace("leave_", ""), []
                    ).append(method)
            self._batched_transformer = _BatchedTransformer(leave_methods)

        self._batched_transformer.update_children_context(self.context)

        logger.debug("Checking {}", self.context.filename)
        was_modified = False
        modified_tree = tree
        for _ in range(self.max_executions):
            self._mark_as_not_modified()
            modified_tree = modified_tree.visit(self._batched_transformer)
            if not self._modified():
                break
            was_modified = True

        # This is a hack to avoid converting an umodified tree to code. This
        # improves the running time by about 10%.
        if not was_modified:
            raise libcst.codemod.SkipFile()
        logger.debug("Modified {}", self.context.filename)
        return modified_tree

    @contextlib.contextmanager
    def _handle_metadata_reference(
        self, module: libcst.Module
    ) -> Generator[libcst.Module, None, None]:
        """Optimize the speed of the orginal _handle_metadata_reference.

        Given that we know that CraftierTransform generate always different
        nodes, we can avoid the copy of the whole tree when building the
        metadata. This is an important performance optimization.
        """
        oldwrapper = self.context.wrapper
        metadata_manager = self.context.metadata_manager
        filename = self.context.filename
        if metadata_manager and filename:
            # We can look up full-repo metadata for this codemod!
            cache = metadata_manager.get_cache_for_path(filename)
            wrapper = libcst.metadata.MetadataWrapper(
                module, cache=cache, unsafe_skip_copy=True
            )
        else:
            # We are missing either the repo manager or the current path,
            # which can happen when we are codemodding from stdin or when
            # an upstream dependency manually instantiates us.
            wrapper = libcst.metadata.MetadataWrapper(
                module, unsafe_skip_copy=True
            )

        with self.resolve(wrapper):
            self.context = dataclasses.replace(self.context, wrapper=wrapper)
            try:
                yield wrapper.module
            finally:
                self.context = dataclasses.replace(
                    self.context, wrapper=oldwrapper
                )

    def _mark_as_not_modified(self) -> None:
        context = self.context.scratch.setdefault(CONTEXT_KEY, {})
        context[self.context.filename] = False

    def _modified(self) -> bool:
        context = self.context.scratch.get(CONTEXT_KEY, {})
        return bool(context[self.context.filename])


class _BatchedTransformer(libcst.CSTTransformer):
    def __init__(
        self,
        leave_methods: MutableMapping[str, List[_LeaveMethod]],
    ):
        libcst.CSTTransformer.__init__(self)
        self.leave_methods = leave_methods

    def on_visit(self, node: libcst.CSTNode) -> bool:
        return True

    def on_leave(
        self, original_node: libcst.CSTNodeT, updated_node: libcst.CSTNodeT
    ) -> Union[libcst.CSTNodeT, libcst.RemovalSentinel]:
        new_updated_node: Union[
            libcst.CSTNodeT, libcst.RemovalSentinel
        ] = updated_node
        original_node_type = type(original_node)
        for on_leave in self.leave_methods.get(original_node_type.__name__, []):
            # We use type here to detect whether the returned node is still
            # processable by these methods.
            # pylint: disable=unidiomatic-typecheck
            if type(updated_node) != original_node_type:
                break
            new_updated_node = on_leave(original_node, new_updated_node)
            # pylint: enable=unidiomatic-typecheck

        return new_updated_node

    def on_visit_attribute(self, node: libcst.CSTNode, attribute: str) -> None:
        return None

    def on_leave_attribute(
        self, original_node: libcst.CSTNode, attribute: str
    ) -> None:
        return None

    def update_children_context(
        self, context: libcst.codemod.CodemodContext
    ) -> None:
        """Propagate the context of the codemod to all children.

        There is some sort of a hack in the LibCST CLI interfaces which change
        the context of an existing codemod, but that does not get propagated all
        the way down.
        """
        for methods in self.leave_methods.values():
            for method in methods:
                method.__self__.context = context



import importlib
import inspect
import pkgutil
import types
from typing import Any, Iterator, Sequence, Set, Type, TypeVar


class InvalidName(Exception):
    """Exception raised when a package or module cannot be imported."""

    def __init__(self, package_name: str):
        super().__init__(f"{package_name} was not found")
        self.package_name = package_name


T = TypeVar("T", bound=Type[Any])


def from_qualified_names(
    base_class: T, qualified_names: Sequence[str]
) -> Sequence[T]:
    """Find all the subclasses from the target class in the given packages."""
    classes: Set[T] = set()
    for name in qualified_names:
        classes.update(_from_package_name(base_class, name))
    return sorted(classes, key=lambda class_: class_.__name__)


def _from_package_name(base_class: T, qualified_name: str) -> Iterator[T]:
    try:
        package = importlib.import_module(qualified_name)
    except ModuleNotFoundError:
        raise InvalidName(qualified_name) from None

    yield from _from_module(base_class, package)

    # Typeshed is missing __path__ for ModuleType. See
    # https://github.com/python/typeshed/issues/4812
    package_path = getattr(package, "__path__", None)
    if not package_path:
        return
    for module_info in pkgutil.iter_modules(
        package_path, prefix=f"{qualified_name}."
    ):
        try:
            submodule = importlib.import_module(module_info.name)
        except ModuleNotFoundError:
            continue
        yield from _from_module(base_class, submodule)


def _from_module(base_class: T, module: types.ModuleType) -> Iterator[T]:
    for _, class_ in inspect.getmembers(module, inspect.isclass):
        if issubclass(class_, base_class):
            yield class_



import dataclasses
from typing import Sequence

import libcst

_SKIP_TO_STRING = frozenset(
    (
        "lbrace",
        "rbrace",
        "lpar",
        "rpar",
        "comma",
        "semicolon",
        "header",
        "footer",
    )
)


def to_string(node: libcst.CSTNode) -> str:
    """Yaml-like representation of a node."""
    fields = [
        f
        for f in dataclasses.fields(node)
        if f.name not in _SKIP_TO_STRING
        and "whitespace" not in f.name
        and not f.name.startswith("_")
    ]

    if not fields:
        return f"{type(node).__name__}"

    lines = [f"{type(node).__name__}:"]
    for f in fields:
        key = f.name
        value = getattr(node, key)
        if value == []:
            value = ()
        if f.default != value:
            value_repr = _pretty_repr(value)
            if not value_repr.startswith("\n"):
                value_repr = f" {value_repr}"
            lines.append(_indent(f"{key}:{value_repr}"))
    return "\n".join(lines)


def _indent(value: str) -> str:
    return "\n".join(f"  {l}" for l in value.split("\n"))


def _pretty_repr(value: object) -> str:
    if isinstance(value, libcst.CSTNode):
        return to_string(value)

    if not isinstance(value, str) and isinstance(value, Sequence):
        return _pretty_repr_sequence(value)

    return repr(value)


def _pretty_repr_sequence(seq: Sequence[object]) -> str:
    if len(seq) == 0:
        return "[]"

    return "\n".join(["", *[f"- {_pretty_repr(el)}" for el in seq]])



import dataclasses
import itertools
import time
import types
from typing import Any, Callable, Dict, Optional, cast

import libcst
import libcst.codemod
import libcst.matchers
import libcst.metadata
from loguru import logger

import craftier.matcher
from craftier import codemod, function_parser, parenthesize, performance


@dataclasses.dataclass
class ExpressionTransform:
    """Represent all the aspects of a code transform."""

    before: libcst.matchers.BaseMatcherNode
    after: libcst.CSTNode
    wrapper: libcst.metadata.MetadataWrapper

    def visitor_method_name(self) -> str:
        """Return the name of the vistior method that could match the matcher.

        For example if the before matcher was a `x + 2`, then the name returned
        would be `leave_BinaryOperation`.
        """
        node_type = type(self.before).__name__
        return f"leave_{node_type}"


def _get_expression_transform(
    before: Callable[..., Any], after: Callable[..., Any]
) -> ExpressionTransform:
    expression = function_parser.parse(before)[0]
    matchers = function_parser.args_to_matchers(before)
    matcher = craftier.matcher.from_node(expression, matchers)
    inner_matcher = getattr(matcher, "matcher", None)
    if isinstance(matcher, libcst.matchers.DoNotCareSentinel) or isinstance(
        inner_matcher, libcst.matchers.DoNotCareSentinel
    ):
        raise Exception(
            f"DoNotCare matcher is forbidden at top level in `{before.__name__}`"
        )

    after_expression = function_parser.parse(after)[0]
    # Technically this is not correct as some expressions,like function calls,
    # binary operations, name, etc, are wrapped in an `Expr` node.
    module = libcst.Module(
        body=[
            libcst.SimpleStatementLine(
                body=[cast(libcst.BaseSmallStatement, after_expression)]
            )
        ]
    )
    wrapper = libcst.metadata.MetadataWrapper(module)
    body = cast(libcst.SimpleStatementLine, wrapper.module.body[0])
    replacement = body.body[0]

    return ExpressionTransform(
        before=matcher, after=replacement, wrapper=wrapper
    )


def check_matches(
    matches: Dict[str, libcst.CSTNode]
) -> Optional[Dict[str, libcst.CSTNode]]:
    """Check all the placeholders matches are sound between them.

    When using placeholders in a node matcher, those are replaced by a special
    name of the form <name>/<occurrence>. They can match any nodes, but at the
    end we need to make sure matches for placeholders with the same name are
    equivalent.

    TODO: instead of calling this method directly, we should incorporate it
    directly into the returned global matcher.
    """
    result = {}
    for key, grouped_matches in itertools.groupby(
        sorted(matches.items()), lambda x: x[0].split("/", 1)[0]
    ):
        first_match, *remaining_matches = grouped_matches
        result[key] = first_match[1]
        if not remaining_matches:
            continue
        matcher = craftier.matcher.from_node(first_match[1], {})
        if not all(
            craftier.matcher.matches(match[1], matcher)
            for match in remaining_matches
        ):
            return None
    return result


class _ReplaceTransformer(libcst.CSTTransformer):
    METADATA_DEPENDENCIES = (libcst.metadata.ParentNodeProvider,)

    def __init__(self, replacements: Dict[str, libcst.CSTNode]) -> None:
        super().__init__()
        self.replacements = replacements

    def leave_Name(
        self, original_node: libcst.Name, updated_node: libcst.Name
    ) -> libcst.BaseExpression:
        new_node = self.replacements.get(updated_node.value)
        if not new_node:
            return updated_node
        parent = self.get_metadata(
            libcst.metadata.ParentNodeProvider, original_node
        )
        if not parent:
            raise Exception("cannot find parent for node")
        # TODO: check if we need to clone the node before returning it. It may
        # have some implications for the MetadataWrapper, as we are forcing to
        # not copy the tree.
        return parenthesize.parenthesize_using_parent(
            cast(libcst.BaseExpression, new_node), parent
        )


def _replace_names(
    node: libcst.CSTNode,
    wrapper: libcst.metadata.MetadataWrapper,
    replacements: Dict[str, libcst.CSTNode],
) -> libcst.CSTNode:
    replacer = _ReplaceTransformer(replacements)
    with replacer.resolve(wrapper):
        # The result of node.visit can never be a RemovalSentinel.
        return cast(libcst.CSTNode, node.visit(replacer))


class CraftierTransformer(codemod.ContextAwareTransformer):
    """Base class for all Craftier based transformers.

    This will correctly generate the appropiate leave methods based on the code
    inside the methods `{check}_before` and `{check}_after`. Note that you must
    specify a pair of such methods.

    For example, if you wanted to replace `None` equality checks with `is`
    checks, you could write something like::

        class NoneEqualityTransformer(CraftierTransformer):
            def equality_before(self, x):
                x == None

            def equality_after(self, x):
                x is None

    The methods `_before` and `_after` need to specify as arguments the
    parametric replacements and will match anything by default. If you want to
    restrict what a given parameter can match, you can use a type annotation as
    follows::

        x: Annotated[Any, matcher]

    where `matcher` is any instance of a valid `libcst.matcher`.

    There are also some predefined types aliases in `craftier.matchers`, for
    example `BooleanLiteral`, which would only match nodes `True` and `False`.

    At the moment, the actual type annoation (first argument of `Annotated`) is
    not used at all, but in the future we plan to leverage `libcst`'s typing
    metadata.
    """

    def __init__(self, context: libcst.codemod.CodemodContext):
        super().__init__(context)
        method_names = set()
        for name in type(self).__dict__:
            if name.endswith("_before") or name.endswith("_after"):
                method_names.add(name.rsplit("_", 1)[0])
        if len(method_names) != 1:
            raise Exception(
                f"Only 1 method is supported at the moment. Found: {method_names}"
            )
        # TODO: handle more method definitions
        name = list(method_names)[0]
        before = getattr(self, f"{name}_before", None)
        after = getattr(self, f"{name}_after", None)
        if not before or not after:
            raise Exception(
                f"Expected `{name}_before` and `{name}_after` to be defined"
            )
        transform = _get_expression_transform(before, after)

        perf_name = f"{self.__class__.__name__}/{name}"
        # pylint: disable=unused-argument
        def custom_on_leave(
            self: CraftierTransformer,
            original_node: libcst.CSTNode,
            updated_node: libcst.CSTNode,
        ) -> libcst.CSTNode:
            new_node = updated_node
            status = "unchanged"
            start = time.perf_counter_ns()
            # We don't support wildcard patterns yet, so we can safely cast the
            # results to single nodes.
            matches = cast(
                Dict[str, libcst.CSTNode],
                libcst.matchers.extract(updated_node, transform.before),
            )
            if matches is not None:
                actual_matches = check_matches(matches)
                if actual_matches is not None:
                    # TODO: make sure all new nodes have a different id
                    new_node = _replace_names(
                        transform.after, transform.wrapper, actual_matches
                    )
                    logger.debug(
                        "{} changed file {}",
                        perf_name,
                        self.context.filename,
                    )
                    new_node = parenthesize.parenthesize_using_previous(
                        new_node, updated_node
                    )
                    status = "modified"
                    self._mark_as_modified()
            end = time.perf_counter_ns()
            performance.write(
                {
                    "name": perf_name,
                    "time": (end - start) / 1e6,
                    "status": status,
                }
            )
            return new_node

        # pylint: enable=unused-argument

        custom_on_leave.__name__ = transform.visitor_method_name()
        bound_leave = types.MethodType(custom_on_leave, self)
        setattr(self, transform.visitor_method_name(), bound_leave)

    def _mark_as_modified(self) -> None:
        context = self.context.scratch.setdefault(codemod.CONTEXT_KEY, {})
        context[self.context.filename] = True



import inspect
import textwrap
from typing import Any, Callable, Dict, Sequence

import libcst
import libcst.matchers
import typing_extensions
from typing_extensions import Annotated

from craftier import libcst_typing


def args_to_matchers(
    function: Callable[..., Any],
) -> Dict[str, libcst_typing.Matcher]:
    """Extract node matchers from the function arguments.

    Untyped arguments will get a `DoNotCare` matcher, while arguments typed and
    annoated with a `BaseMatcherNode` will return that matcher.
    """
    matchers: Dict[str, libcst_typing.Matcher] = {}

    # Create default matchers for all arguments
    args = function.__code__.co_varnames[: function.__code__.co_argcount]
    if args:
        for name in args:
            if name == "self":
                continue
            matchers[name] = libcst.matchers.DoNotCare()

    # Check if any of the arguments was annotated with a Matcher
    for name, type_declaration in typing_extensions.get_type_hints(
        function, include_extras=True
    ).items():
        if typing_extensions.get_origin(type_declaration) is Annotated:
            # Check if there is a matcher
            arg = typing_extensions.get_args(type_declaration)[1]
            if isinstance(arg, libcst.matchers.BaseMatcherNode):
                matchers[name] = arg

    return matchers


class UnusedArgument(Exception):
    """Exception raised when an argument is unused.

    Attributes:
      name: the name of the unused argument.
      function_name: the name of the function where the argument is defined.
    """

    name: str
    function_name: str

    def __init__(self, name: str, fn_name: str, lineno: int) -> None:
        super().__init__(
            f'Unused argument "{name}" in function {fn_name}:{lineno}'
        )
        self.name = name
        self.fn_name = fn_name
        self.lineno = lineno


def _remove_expression(node: libcst.CSTNode) -> libcst.CSTNode:
    if isinstance(node, libcst.Expr):
        return node.value
    return node


def _from_body(body: Sequence[libcst.CSTNode]) -> Sequence[libcst.CSTNode]:
    if len(body) == 1:
        stmt = body[0]
        if isinstance(
            stmt, (libcst.SimpleStatementLine, libcst.SimpleStatementSuite)
        ):
            if len(stmt.body) == 1:
                return [_remove_expression(stmt.body[0])]
        if isinstance(stmt, libcst.Expr):
            return [_remove_expression(stmt)]
    return body


def _from_function_source(
    source: str, function: Callable[..., Any]
) -> Sequence[libcst.CSTNode]:
    source = textwrap.dedent(source)
    tree = libcst.parse_statement(source)
    function_node = libcst.ensure_type(tree, libcst.FunctionDef)
    body = function_node.body.body
    if function_node.get_docstring():
        body = body[1:]

    args = function.__code__.co_varnames[: function.__code__.co_argcount]
    if args:
        for name in args:
            if name == "self":
                continue
            if not libcst.matchers.findall(
                function_node.body, libcst.matchers.Name(value=name)
            ):
                raise UnusedArgument(
                    name, function.__name__, function.__code__.co_firstlineno
                )

    return _from_body(body)


def parse(function: Callable[..., Any]) -> Sequence[libcst.CSTNode]:
    """Extract a `CSTNode` from a function's source code."""
    return _from_function_source(inspect.getsource(function), function)



import pathlib
from typing import Iterable, Sequence


def get_modified_files(
    files: Iterable[pathlib.Path], *, since: float
) -> Sequence[pathlib.Path]:
    """Return the files that were modified since the given timestamp."""
    return [f for f in files if f.stat().st_mtime > since]



import collections
import multiprocessing
import pathlib
import platform
import statistics
import sys
from typing import Any, DefaultDict, List, Mapping, Optional, Sequence

import click
import click_pathlib
from loguru import logger

from craftier import config, performance, refactor

# Generated using
# http://patorjk.com/software/taag/#p=display&f=Graceful&t=craftier
# The \b signals that paragraph won't be rewrapped.
_BANNER = (
    "\b"
    + r"""
  ___  ____   __   ____  ____  __  ____  ____ (pre-α)
 / __)(  _ \ / _\ (  __)(_  _)(  )(  __)(  _ \
( (__  )   //    \ ) _)   )(   )(  ) _)  )   /
 \___)(__\_)\_/\_/(__)   (__) (__)(____)(__\_)

      Your personal Python code reviewer
    """
)

_CONTEXT_SETTINGS = {"help_option_names": ["-h", "--help"]}


@click.group(context_settings=_CONTEXT_SETTINGS, help=_BANNER)
@click.version_option(version="0.1.0", prog_name="craftier")
def app() -> None:
    """Entry point for craftier CLI."""


@click.command(
    context_settings=_CONTEXT_SETTINGS,
    name="refactor",
    help="Refactor your code",
)
@click.argument(
    "files",
    nargs=-1,
    type=click_pathlib.Path(
        exists=True,
        writable=True,
        readable=True,
        dir_okay=False,
        allow_dash=True,
    ),
)
@click.option(
    "--config",
    "config_path",
    type=click_pathlib.Path(exists=True, readable=True, dir_okay=False),
    help="Path to the config file. Searches for .craftier.ini by default.",
)
@click.option(
    "--debug/--no-debug",
    default=False,
    is_flag=True,
    help="Show extra debugging information.",
)
def _refactor(
    files: Sequence[pathlib.Path],
    config_path: Optional[pathlib.Path],
    debug: bool,
) -> None:
    """Subcommand to refactor code."""
    # In Python 3.8 for Mac OS the default was changed from fork to spawn,
    # however that brings lot of undesired effects. For example the Pickling
    # of many structures, which is fragile and the errors are undecipherable.
    # Aditionally it means, that logging and global state won't be preserved.
    # Spawning is also slower.
    # Note that it should be safe to use fork as we don't have any
    # multi-threaded code.
    if (
        platform.system() == "Darwin"
        and multiprocessing.get_start_method(allow_none=True) != "fork"
    ):
        multiprocessing.set_start_method("fork")
    level = "DEBUG" if debug else "WARNING"
    logger.remove()
    logger.add(
        sys.stderr,
        level=level,
        format="{level}\t{elapsed}\t{message}",
        enqueue=True,
    )
    logger.enable("craftier")

    if debug:
        performance.enable()

    if not files:
        click.secho("No path provided. Nothing to do 💤", bold=True)
        return
    try:
        craftier_config = config.load(config_path)
        logger.debug("Config path: {}", craftier_config.path)
        result = refactor.run(craftier_config, files)
        _fix_summary(result)
    except (config.InvalidConfigError, refactor.Error) as e:
        print(e)
        raise click.exceptions.Exit(1)
    except KeyboardInterrupt:
        raise click.Abort("Aborted by user") from None

    if debug:
        performance.disable()


def _fix_summary(result: refactor.Result) -> None:
    _print_performance_stats()

    for file in result.refactored:
        click.secho(f"refactored {file.relative_to('.')}", bold=True)

    if result.success:
        click.secho("All done! 🎉 🏆 🎉", bold=True)
    else:
        click.secho("Oh no! 🧨 💣 🧨", bold=True)
    refactored = len(result.refactored)
    unchanged = len(result.unchanged)
    errors = result.error_count
    report = []
    if refactored > 0:
        report.append(
            click.style(
                f"{refactored} {_pluralize('file', refactored)} refactored",
                bold=True,
            )
        )
    if unchanged > 0:
        report.append(
            f"{unchanged} {_pluralize('file', unchanged)} left unchanged"
        )
    if errors > 0:
        report.append(
            click.style(
                f"{errors} {_pluralize('file', errors)} failed", fg="red"
            )
        )
    click.echo(f"{', '.join(report)}.")
    if not result.success:
        raise click.exceptions.Exit(1)


def _pluralize(word: str, count: int) -> str:
    ending = "s" if count != 1 else ""
    return f"{word}{ending}"


def _print_performance_stats() -> None:
    data = performance.read()
    grouped_data: DefaultDict[
        str, List[Mapping[str, Any]]
    ] = collections.defaultdict(list)
    for row in data:
        grouped_data[row["name"]].append(row)
    for name in sorted(grouped_data.keys()):
        _print_single_performance_stats(name, grouped_data[name])
    _print_single_performance_stats("Overall", data)


def _print_single_performance_stats(
    name: str, data: Sequence[Mapping[str, Any]]
) -> None:
    modified = [row["time"] for row in data if row["status"] == "modified"]
    unchanged = [row["time"] for row in data if row["status"] == "unchanged"]
    values = [row["time"] for row in data]
    logger.debug(
        "{}\tcount={}\ttotal={:.1f}ms\t{}\t{}\t{}",
        name,
        len(values),
        sum(values),
        _stats(values),
        _stats(modified),
        _stats(unchanged),
    )


def _stats(data: List[float]) -> str:
    if not data:
        return "-/-/-"
    mean = statistics.mean(data)
    maximum = max(data)
    try:
        stddev = statistics.stdev(data, xbar=mean)
    except statistics.StatisticsError:
        stddev = 0
    return f"μ={mean:.1f}ms/σ={stddev:.2f}ms/max={maximum:.1f}ms"


app.add_command(_refactor)



import ast
import dataclasses
from typing import Any, Callable, Dict, List, Sequence, Tuple, Union, cast

import libcst
import libcst.matchers

from craftier import libcst_typing

# TODO: have a custom extension of OneOf and MatchIfTrue, so that we can know
# which classes to match on.
#
# Could be something like:
#
# class MatchIfTrue extends libcst.matchers.MatchIfTrue:
#     __init__(self, expected_types, condition):
#         self.expected_types
#         def new_condition(node):
#             if not isinstance(node, expected_types):
#                 return false
#             return condition(node)
#         super()(new_condition)
#
# class OneOf extends libcst.matchers.OneOf:
#     property to extract all expected types to listen to

# TODO: add special matcher for qualified names: for example math.log should
# match log or renames as well.
# TODO: preserve leading_lines, this is required to preserve previous comments.
# Comments inside replaced code are going to be removed
# TODO: assign should match annassign


@dataclasses.dataclass
class _Placeholder:
    """Helper class used to keep track how many matchers are in use.

    When we are building a matcher from sorce code, we don't know how many times
    we have seen that placeholder. So in order to track that we wrap the
    matchers in this counter class.

    In that way we can give the instances of pattern x the names x/0, x/1, x/2
    and so forth. Then after a match is found we will check that indeed all
    instances of x/0, x/1, etc. are equivalent.
    """

    matcher: libcst_typing.Matcher
    total: int = 0


_SKIP_FIELDS = frozenset(
    (
        "lbrace",
        "rbrace",
        "lpar",
        "rpar",
        "comma",
        "semicolon",
        "header",
        "footer",
        "first_colon",
        "second_colon",
    )
)


def _skip_field(name: str) -> bool:
    if name in _SKIP_FIELDS:
        return True
    return "whitespace" in name or name.startswith("_")


def _repr_single(string: str) -> str:
    # See https://stackoverflow.com/questions/27402168/force-repr-to-use-single-quotes
    return "'" + repr('"' + string)[2:]


def _make_literal_value_matcher(
    node_value: str,
) -> libcst.matchers.MatchIfTrue[Callable[[str], bool]]:
    expected_value = ast.literal_eval(node_value)

    def comparator(value: str) -> bool:
        return cast(bool, ast.literal_eval(value) == expected_value)

    return libcst.matchers.MatchIfTrue(comparator)


def _make_simple_string_matcher(
    original_node: libcst.SimpleString,
) -> libcst.matchers.BaseMatcherNode:
    original_value = original_node.evaluated_value

    def concatenated_condition(node: libcst.CSTNode) -> bool:
        if not isinstance(node, libcst.ConcatenatedString):
            return False
        flattened_node = _flatten_concatenated_string(node)
        return (
            isinstance(flattened_node, libcst.SimpleString)
            and flattened_node.evaluated_value == original_value
        )

    concatenated_matcher = libcst.matchers.MatchIfTrue(concatenated_condition)
    return libcst.matchers.OneOf(
        libcst.matchers.SimpleString(
            value=_make_literal_value_matcher(original_node.value)
        ),
        concatenated_matcher,
    )


def _formatted_string_parts(
    node: libcst.FormattedString,
) -> Tuple[Sequence[str], Sequence[libcst.FormattedStringExpression]]:
    texts = []
    expressions = []
    prefix = node.start.replace("f", "").replace("F", "")
    for part in node.parts:
        if isinstance(part, libcst.FormattedStringText):
            texts.append(
                cast(str, ast.literal_eval(f"{prefix}{part.value}{node.end}"))
            )
        elif isinstance(part, libcst.FormattedStringExpression):
            expressions.append(part)
    return texts, expressions


def _make_formatted_string_matcher(
    original_node: libcst.FormattedString, placeholders: Dict[str, _Placeholder]
) -> libcst.matchers.BaseMatcherNode:
    original_texts, original_expressions = _formatted_string_parts(
        original_node
    )
    original_expressions_matchers = [
        _to_matcher(e, placeholders) for e in original_expressions
    ]

    def formatted_condition(node: libcst.CSTNode) -> bool:
        if isinstance(node, libcst.ConcatenatedString):
            node = _flatten_concatenated_string(node)

        if not isinstance(node, libcst.FormattedString):
            return False

        texts, expressions = _formatted_string_parts(node)
        if texts != original_texts or len(expressions) != len(
            original_expressions_matchers
        ):
            return False
        return all(
            matches(e, matcher)
            for e, matcher in zip(expressions, original_expressions_matchers)
        )

    # We need to wrap it with OneOf, otherwise it won't match
    return libcst.matchers.OneOf(
        libcst.matchers.MatchIfTrue(formatted_condition)
    )


def _flatten_concatenated_string(
    node: libcst.ConcatenatedString,
) -> Union[libcst.SimpleString, libcst.FormattedString]:
    classes = {type(node.left)}
    parts = []
    rest: Union[
        libcst.ConcatenatedString, libcst.SimpleString, libcst.FormattedString
    ] = node
    while isinstance(rest, libcst.ConcatenatedString):
        parts.append(rest.left)
        classes.add(type(rest.left))
        rest = rest.right
    parts.append(rest)
    classes.add(type(rest))
    # print(parts)

    if all(isinstance(n, libcst.SimpleString) for n in parts):
        # There's no idiom other than casting to tell mypy the list only has one
        # of the union elements. See https://github.com/python/mypy/issues/3497
        string_parts = cast(List[libcst.SimpleString], parts)
        content = "".join(n.evaluated_value for n in string_parts)
        return libcst.SimpleString(value=repr(content))

    formatted_parts: List[libcst.BaseFormattedStringContent] = []
    for part in parts:
        if isinstance(part, libcst.SimpleString):
            value = _repr_single(part.evaluated_value)[1:-1]
            if formatted_parts and isinstance(
                formatted_parts[-1], libcst.FormattedStringText
            ):
                formatted_parts[-1] = libcst.FormattedStringText(
                    value=formatted_parts[-1].value + value
                )
            else:
                formatted_parts.append(libcst.FormattedStringText(value=value))
        else:
            for nested_part in part.parts:
                prefix = part.start.replace("f", "").replace("F", "")
                if isinstance(nested_part, libcst.FormattedStringText):
                    value = _repr_single(
                        ast.literal_eval(
                            f"{prefix}{nested_part.value}{part.end}"
                        )
                    )[1:-1]
                    if formatted_parts and isinstance(
                        formatted_parts[-1], libcst.FormattedStringText
                    ):
                        formatted_parts[-1] = libcst.FormattedStringText(
                            value=formatted_parts[-1].value + value
                        )
                    else:
                        formatted_parts.append(
                            libcst.FormattedStringText(value=value)
                        )
                else:
                    formatted_parts.append(nested_part)
    return libcst.FormattedString(parts=formatted_parts, start="f'", end="'")


def _to_matcher(
    node: libcst.CSTNode, placeholders: Dict[str, _Placeholder]
) -> libcst_typing.Matcher:
    if isinstance(node, libcst.Name) and node.value in placeholders:
        placeholder = placeholders[node.value]
        placeholder.total += 1
        return libcst.matchers.SaveMatchedNode(
            placeholder.matcher, f"{node.value}/{placeholder.total}"
        )

    if isinstance(node, (libcst.Integer, libcst.Float, libcst.Imaginary)):
        return cast(
            libcst.matchers.BaseMatcherNode,
            getattr(libcst.matchers, node.__class__.__name__)(
                value=_make_literal_value_matcher(node.value)
            ),
        )

    if isinstance(node, libcst.ConcatenatedString):
        node = _flatten_concatenated_string(node)

    if isinstance(node, libcst.SimpleString):
        return _make_simple_string_matcher(node)

    if isinstance(node, libcst.FormattedString):
        return _make_formatted_string_matcher(node, placeholders)

    fields = [f for f in dataclasses.fields(node) if not _skip_field(f.name)]

    data: Dict[str, Any] = {}
    matcher: Any
    for f in fields:
        key = f.name
        value = getattr(node, key)
        # print(key, value, type(value))
        # if value == []:
        #    value = ()
        # if f.default != value:
        #    lines.append(_indent(f"{key}: {_pretty_repr(value)}"))
        if isinstance(value, libcst.CSTNode):
            matcher = _to_matcher(value, placeholders)
        elif not isinstance(value, str) and isinstance(value, Sequence):
            if value and isinstance(value[0], libcst.CSTNode):
                matcher = [_to_matcher(n, placeholders) for n in value]
            else:
                matcher = value
        elif value == libcst.MaybeSentinel.DEFAULT:
            matcher = libcst.matchers.DoNotCare()
        else:
            matcher = value
        # if key == 'operator':
        #    print(matcher, type(matcher))
        data[key] = matcher
    # print('matcher class', getattr(m, node.__class__.__name__))
    return cast(
        libcst.matchers.BaseMatcherNode,
        getattr(libcst.matchers, node.__class__.__name__)(**data),
    )


def from_node(
    node: libcst.CSTNode, placeholders: Dict[str, libcst_typing.Matcher]
) -> libcst_typing.Matcher:
    """Return a matcher that matches the given node.

    The matcher will be semantic, meaning that equivalent nodes will be matched.
    For example: literal nodes are compared by value and strings are simplified.

    The placeholders mapping specifies how to match certain names. For example
    if you pass {'x': DoNotCare()}, then the node `x + 1`, will match any
    expression that's being added to 1.

    Args:
      node: the node to generate a matcher for
      placeholders: mapping of node name to matcher
    """
    wrapped_placeholders = {
        p: _Placeholder(matcher) for p, matcher in placeholders.items()
    }
    # TODO: return a compound matcher that checks that multiple instances of the
    # same placeholder matches.
    # Cool thing is that if there's no repeated placeholer, we will incur no
    # extra cost while checking.
    # This will be an AllOf matcher. We just need to make sure that we return
    # the right matcher type.
    return _to_matcher(node, wrapped_placeholders)


def matches(node: libcst.CSTNode, matcher: libcst_typing.Matcher) -> bool:
    """Return True if the node matches the shape defined by the matcher.

    This is a replacement for libcst.matchers.matches accepting a
    DoNotCareSentinel.
    """
    if isinstance(matcher, libcst.matchers.DoNotCareSentinel):
        return True
    return libcst.matchers.matches(node, matcher)



import contextlib
import dataclasses
import io
import os
import pathlib
import platform
import sys
import time
from typing import Any, Sequence, Type

import libcst
import libcst.codemod
from loguru import logger

from craftier import class_finder, codemod
from craftier import config as craftier_config
from craftier import fs

_LIBCST_SINK = io.StringIO()


class Error(Exception):
    """Base class for refactoring errors."""


class EmptyTransformersError(Error):
    """Error signaling no transformers were applied."""

    def __init__(self) -> None:
        super().__init__("No tranformers were specified")


class InvalidPackageError(Error):
    """Error signaling one or more package names are invalid."""

    def __init__(self, package_name: str) -> None:
        super().__init__(
            f"Package {package_name} is not valid. Are you missing any dependencies?"
        )
        self.package_name = package_name


@dataclasses.dataclass(frozen=True)
class Result:
    """Summary data of a craftier run command."""

    refactored: Sequence[pathlib.Path]
    unchanged: Sequence[pathlib.Path]
    success: bool
    error_count: int


def run(
    config: craftier_config.Config, files: Sequence[pathlib.Path]
) -> Result:
    """Execute craftier for the given set of files.

    Raises:
      EmptyTransformersError: when no transformers are selected.
      InvalidPackageError: when packages are missing or invalid.
    """
    try:
        transformers = _get_transformers(config)
    except class_finder.InvalidName as e:
        raise InvalidPackageError(e.package_name) from None
    if not transformers:
        raise EmptyTransformersError()
    logger.opt(lazy=True).debug(
        "Transformers: {}",
        lambda: ", ".join(_full_name(t) for t in transformers),
    )

    transform = codemod.BatchedCodemod(
        libcst.codemod.CodemodContext(), transformers
    )
    python_files = [str(p) for p in files]
    now = time.time()
    result = _exec_transform(transform, python_files)

    modified = fs.get_modified_files(files, since=now)

    if result.failures:
        errors = _LIBCST_SINK.getvalue()
        if errors:
            print("\n\nLibCST raw logs", file=sys.stderr)
            print("-" * 80, file=sys.stderr)
            print(errors, file=sys.stderr)
            print("-" * 80, file=sys.stderr)

    return Result(
        refactored=modified,
        unchanged=[p for p in files if p not in modified],
        success=result.failures == 0,
        error_count=result.failures,
    )


def _full_name(class_: Type[Any]) -> str:
    return f"{class_.__module__}.{class_.__qualname__}"


def _get_transformers(
    config: craftier_config.Config,
) -> Sequence[Type[codemod.ContextAwareTransformer]]:
    excluded = set(config.excluded)
    transformers = class_finder.from_qualified_names(
        codemod.ContextAwareTransformer, config.packages
    )
    return [t for t in transformers if _full_name(t) not in excluded]


def _is_profiling() -> bool:
    """Detect whether the code is being profiled.

    Note: It only detects whther the code is being profiled under cProfile.
    """
    return "_lsprof" in sys.modules


def _is_windows() -> bool:
    # See https://stackoverflow.com/a/7637706/1413687
    system = platform.system()
    return system == "Windows" or system.startswith("CYGWIN_NT")


def _sequential_exec_transform_with_prettyprint(
    transform: libcst.codemod.Codemod,
    python_files: Sequence[str],
    *,
    include_generated: bool = True,
) -> libcst.codemod.ParallelTransformResult:
    """Run a transform sequentially in the same process.

    This wrapper is similar in API to parallel_exec_transform_with_prettyprint
    and it is used when running the code on Windows or under a profiler.
    """
    for file in python_files:
        transform.context = dataclasses.replace(
            transform.context, filename=file
        )
        with open(file) as f:
            libcst.codemod.exec_transform_with_prettyprint(
                transform, f.read(), include_generated=include_generated
            )

    return libcst.codemod.ParallelTransformResult(
        successes=0, failures=0, skips=0, warnings=0
    )


def _exec_transform(
    transform: libcst.codemod.Codemod, files: Sequence[str]
) -> libcst.codemod.ParallelTransformResult:
    if _is_profiling() or _is_windows():
        return _sequential_exec_transform_with_prettyprint(
            transform, files, include_generated=True
        )

    chunksize = 4
    jobs = min(
        os.cpu_count() or 1,
        (len(files) + chunksize - 1) // chunksize,
    )
    with contextlib.redirect_stderr(_LIBCST_SINK):
        return libcst.codemod.parallel_exec_transform_with_prettyprint(
            transform,
            files,
            include_generated=True,
            show_successes=False,
            jobs=jobs,
        )



from typing import TypeVar

import libcst


def _precedence(node: libcst.CSTNode) -> int:
    """
    Return the precedence of a given expression.

    Nodes which are not expression have the maximum precedence.

    See https://docs.python.org/3/reference/expressions.html for details.
    """
    if isinstance(node, libcst.BinaryOperation):
        binary_operation_precedence = {
            libcst.BitOr: 8,
            libcst.BitXor: 9,
            libcst.BitAnd: 10,
            libcst.LeftShift: 11,
            libcst.RightShift: 11,
            libcst.Add: 12,
            libcst.Subtract: 12,
            libcst.Multiply: 13,
            libcst.MatrixMultiply: 13,
            libcst.Divide: 13,
            libcst.FloorDivide: 13,
            libcst.Modulo: 13,
            libcst.Power: 15,
        }
        return binary_operation_precedence[type(node.operator)]

    if isinstance(node, libcst.BooleanOperation):
        boolean_operation_precedence = {
            libcst.Or: 4,
            libcst.And: 5,
        }
        return boolean_operation_precedence[type(node.operator)]

    if isinstance(node, libcst.UnaryOperation):
        unary_operation_precedence = {
            libcst.Not: 6,
            libcst.BitInvert: 14,
            libcst.Minus: 14,
            libcst.Plus: 14,
        }
        return unary_operation_precedence[type(node.operator)]

    precedence = {
        libcst.NamedExpr: 1,
        libcst.Lambda: 2,
        libcst.IfExp: 3,
        libcst.Comparison: 7,
        libcst.Await: 16,
        libcst.Subscript: 17,
        libcst.Call: 17,
        libcst.Attribute: 17,
        libcst.Tuple: 18,
        libcst.List: 18,
        libcst.ListComp: 18,
        libcst.Dict: 18,
        libcst.DictComp: 18,
        libcst.Set: 18,
        libcst.SetComp: 18,
        libcst.GeneratorExp: 100,
    }
    return precedence.get(type(node), 100)


def _needs_parentheses_parent(
    node: libcst.CSTNode, parent: libcst.CSTNode
) -> bool:
    # The node does not support parenthesis, there's nothing to do
    if not hasattr(node, "lpar"):
        return False

    # The node is already parenthesized, there's nothing to do
    if getattr(node, "lpar"):
        return False

    # A generator expression requires parenthesis, except when it is the sole
    # argument of a function call.
    if isinstance(node, libcst.GeneratorExp):
        if isinstance(parent, libcst.BaseExpression):
            if isinstance(parent, libcst.Call) and len(parent.args) <= 1:
                return False
            return True
        return True

    # Any tuple within an expression requires parenthesis
    if isinstance(node, libcst.Tuple):
        return isinstance(parent, libcst.BaseExpression)

    return _precedence(node) < _precedence(parent)


def _needs_parentheses_previous(
    node: libcst.CSTNode, previous: libcst.CSTNode
) -> bool:
    # The node does not support parenthesis, there's nothing to do
    if not hasattr(node, "lpar"):
        return False

    # The node is already parenthesized, there's nothing to do
    if getattr(node, "lpar"):
        return False

    # The previous node had parentheses, keep them
    if getattr(previous, "lpar", None):
        return True

    # We don't have information from the parent
    if isinstance(node, libcst.GeneratorExp):
        return True

    # We don't have information from the parent
    if isinstance(node, libcst.Tuple):
        return True

    return _precedence(node) < _precedence(previous)


T = TypeVar("T", bound=libcst.CSTNode)


def parenthesize_using_parent(node: T, parent: libcst.CSTNode) -> T:
    """Add parentheses to the given node if needed.

    It will use the parent of the node to decide whether parentheses are
    required.
    """
    if _needs_parentheses_parent(node, parent):
        return node.with_changes(
            lpar=[libcst.LeftParen()], rpar=[libcst.RightParen()]
        )
    return node


def parenthesize_using_previous(node: T, previous: libcst.CSTNode) -> T:
    """Add parentheses to the given node if needed.

    It will use the previous node this node is replacing to decide whether
    parentheses are required.

    Note: this function is not as precise as `parenthesize_using_parent`
    """
    if _needs_parentheses_previous(node, previous):
        return node.with_changes(
            lpar=[libcst.LeftParen()], rpar=[libcst.RightParen()]
        )
    return node



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['craftier', 'craftier.refactors']

package_data = \
{'': ['*']}

install_requires = \
['click>=7.1.2,<8.0.0',
 'click_pathlib>=2020.3.13,<2021.0.0',
 'libcst>=0.3.15,<0.4.0',
 'loguru>=0.5.3,<0.6.0',
 'typing_extensions>=3.7.4,<4.0.0']

entry_points = \
{'console_scripts': ['craftier = craftier.cli:app']}

setup_kwargs = {
    'name': 'craftier',
    'version': '0.1.0',
    'description': '',
    'long_description': '# Craftier\n\n```\n  ___  ____   __   ____  ____  __  ____  ____ (pre-\xce\xb1)\n / __)(  _ \\ / _\\ (  __)(_  _)(  )(  __)(  _ \\\n( (__  )   //    \\ ) _)   )(   )(  ) _)  )   /\n \\___)(__\\_)\\_/\\_/(__)   (__) (__)(____)(__\\_)\n\n      Your personal Python code reviewer\n```\n\n## What is it?\nCraftier is a framework to easily writing Python code refactors. In the near\nfuture it will also come with a set of predefined refactoring rules.\n\nIt is based on `libcst` and simplifies the use of its API by letting you write\nrefactors just by writing Python code.\n\nIt also preserves relevant comments and ensures the modified code is correct, by\nadding required parentheses.\n\n## Getting started\nAfter installing with `pip install craftier`, you can run the default rules with\n`craftier refactor <python files>`.\n\n## Configuration\nBy default we look for a `.craftier.ini`, if none is found a default\nconfiguration will be used.\n\nYou can also specify the config file with `--config CONFIG_PATH`\n\n### Config format\n\n```ini\n[craftier]\npackages=craftier.refactors,\nexcluded=A,\n         B\n```\n\n## Writing your own rules\n\nAs simple as:\n\n```py\nimport craftier\n\nclass SquareTransformer(craftier.CraftierTransformer):\n    def square_before(self, x):\n        x * x\n\n    def square_after(self, x):\n        x ** 2\n\n\nclass IfTrueTransformer(craftier.CraftierTransformer):\n    def if_true_before(self, x, y):\n        x if True else y\n\n    def if_true_after(self, x):\n        x\n```\n\nTODO: write about custom matchers and type declarations\n\n# Roadmap and TODOs\n* Support multiple expressions in a transformer\n* Add support for statements \n* Complete set of refactorings\n* Add support for typing metadata\n* Generate RE2 filtering based on expressions. This could be used to prefilter\n  the list of files and to test the patterns either in codebase search tools\n  like https://grep.app\n* Extensive validation testing\n\n## Limitations\nThis is a work in progress, and some edges are rough.\n\nGiven that we match code based on actual Python code, some refactorings are not\neasily expressed or may not even be expressible at all.\n\n## History\nTODO: write how I came up with the idea.\n\n### Name\nOriginally I wanted to name this package `pythonista`, but unfortunately someone\nsquatted the name (along with several others) and the name was not released\naccording to procedure. Furthermore, the issue was locked.\n\nThat gave me a chance of rethinking the name, so I started with anagrams of the\nword `refactor`, which was of course unsuccessful. So after some\nexperimentation I replaced the letter `o` with an `i`, giving `refactir`, which\nis an anagram of the word `craftier`.\n\nI like that name, because is like a "refactoring" of the word `refactor`, and\n`craftier` conveys the actual use case of the tool.\n\nBonus: the name sounds somewhat like my surname.',
    'author': 'Sebastian Kreft',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/sk-/craftier',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""
Module to keep track of performance statistics across multiple processes.

Note: because this uses a global state it will only work in forked
multi-process. Note that the default changed in Python 38 for MacOs.
"""

import dataclasses
import json
import multiprocessing
import tempfile
from typing import IO, Any, Dict, Mapping, Optional, Sequence


@dataclasses.dataclass
class _Config:
    file: Optional[IO[bytes]] = None


_config = _Config()


class PerformanceError(Exception):
    """Errors specific to the performance module."""


def enable() -> None:
    """Enable the performance sink."""
    if _config.file:
        raise PerformanceError("performance is already enabled")
    _config.file = tempfile.NamedTemporaryFile(
        prefix="craftier-stats-", buffering=0
    )


def write(data: Dict[str, Any]) -> None:
    """Write an entry to the performance file.

    Data must be a JSON serializable object.
    """
    if not _config.file:
        return
    with multiprocessing.Lock():
        _config.file.write(json.dumps(data).encode())
        _config.file.write(b"\n")


def read() -> Sequence[Mapping[str, Any]]:
    """Read the contents of the performance data."""
    if not _config.file:
        return []
    with multiprocessing.Lock():
        _config.file.seek(0)
        return [json.loads(line.strip().decode()) for line in _config.file]


def disable() -> None:
    """Disable the performance sink.

    Further writes and read will fail.
    """
    if not _config.file:
        raise PerformanceError("performance is not enabled")
    _config.file.close()
    _config.file = None



import configparser
import dataclasses
import pathlib
from typing import List, Optional, Sequence, Tuple

# The maximum number of directories to be searched for the config file
_MAX_SEARCH_DEPTH: int = 25
_STOP_SEARCH_ON_DIRS: Tuple[str, ...] = (".git", ".hg")
CONFIG_FILENAME = ".craftier.ini"


class InvalidConfigError(Exception):
    """Error raised when the configuration file is invalid."""


@dataclasses.dataclass
class Config:
    """Craftier's config data.

    It will be typically read from a config file, a .craftier.ini file.
    """

    path: Optional[pathlib.Path]
    packages: Sequence[str] = ("craftier.refactors",)
    excluded: Sequence[str] = ()


def find_path(
    start_path: pathlib.Path = pathlib.Path("."),
) -> Optional[pathlib.Path]:
    """Traverse the file system looking for the config file .craftier.ini.

    It will stop earlier at the user's home directory, if it encounters a Git or
    Mercurial directory, or if it traversed too deep.
    """
    home = pathlib.Path.home()
    path = start_path.resolve()
    for path in [path, *path.parents][:_MAX_SEARCH_DEPTH]:
        config_file = path / CONFIG_FILENAME
        if config_file.is_file():
            return config_file

        for stop_dir in _STOP_SEARCH_ON_DIRS:
            if (path / stop_dir).is_dir():
                return None

        if path == home:
            return None

    return None


def load(path: Optional[pathlib.Path]) -> Config:
    """Load craftier's config.

    It will use the given path or try to detects the config otherwise.
    """
    if not path:
        path = find_path()

    if not path:
        return Config(path=None)

    config = configparser.ConfigParser()
    try:
        config.read(path)
    except configparser.Error as e:
        raise InvalidConfigError(f"{path} is not a valid config file") from e
    try:
        craftier_config = config["craftier"]
    except KeyError as e:
        raise InvalidConfigError("missing [craftier] section") from e

    return Config(
        path=path,
        packages=_parse_list(craftier_config.get("packages", "")),
        excluded=_parse_list(craftier_config.get("excluded", "")),
    )


def _parse_list(value: str) -> List[str]:
    return [part.strip() for part in value.split(",") if part.strip()]



import ast
from typing import Any, Callable, cast

import libcst.matchers
from typing_extensions import Annotated

_BUILTINS = frozenset(
    (
        "abs",
        "all",
        "any",
        "ascii",
        "bin",
        "bool",
        "breakpoint",
        "bytearray",
        "bytes",
        "callable",
        "chr",
        "classmethod",
        "compile",
        "complex",
        "copyright",
        "credits",
        "delattr",
        "dict",
        "dir",
        "divmod",
        "enumerate",
        "eval",
        "exec",
        "exit",
        "filter",
        "float",
        "format",
        "frozenset",
        "getattr",
        "globals",
        "hasattr",
        "hash",
        "help",
        "hex",
        "id",
        "input",
        "int",
        "isinstance",
        "issubclass",
        "iter",
        "len",
        "license",
        "list",
        "locals",
        "map",
        "max",
        "memoryview",
        "min",
        "next",
        "object",
        "oct",
        "open",
        "ord",
        "pow",
        "print",
        "property",
        "quit",
        "range",
        "repr",
        "reversed",
        "round",
        "set",
        "setattr",
        "slice",
        "sorted",
        "staticmethod",
        "str",
        "sum",
        "super",
        "tuple",
        "type",
        "vars",
        "zip",
    )
)


def _is_builtin(value: str) -> bool:
    return value in _BUILTINS


_BOOLEAN_VALUES = frozenset(("True", "False"))


def _is_boolean(value: str) -> bool:
    return value in _BOOLEAN_VALUES


Builtin = Annotated[
    Any, libcst.matchers.Name(value=libcst.matchers.MatchIfTrue(_is_builtin))
]
StringLiteral = Annotated[str, libcst.matchers.SimpleString()]
IntegerLiteral = Annotated[int, libcst.matchers.Integer()]
FloatLiteral = Annotated[float, libcst.matchers.Float()]
BooleanLiteral = Annotated[
    bool, libcst.matchers.Name(value=libcst.matchers.MatchIfTrue(_is_boolean))
]
NoneLiteral = Annotated[None, libcst.matchers.Name(value="None")]

# TODO: export and test these matchers
def _match_if_equal(
    node_value: str,
) -> libcst.matchers.MatchIfTrue[Callable[[str], bool]]:
    expected_value = ast.literal_eval(node_value)

    def comparator(value: str) -> bool:
        return cast(bool, ast.literal_eval(value) == expected_value)

    return libcst.matchers.MatchIfTrue(comparator)


def _match_if_distinct(
    node_value: str,
) -> libcst.matchers.MatchIfTrue[Callable[[str], bool]]:
    expected_value = ast.literal_eval(node_value)

    def comparator(value: str) -> bool:
        return cast(bool, ast.literal_eval(value) != expected_value)

    return libcst.matchers.MatchIfTrue(comparator)


Falsey = Annotated[
    Any,
    libcst.matchers.OneOf(
        libcst.matchers.Name(value="False"),
        libcst.matchers.Integer(value=_match_if_equal("0")),
        libcst.matchers.Float(value=_match_if_equal("0.0")),
        libcst.matchers.SimpleString(value=_match_if_equal("''")),
        # TODO: get matchers for [], (,), {}, check what happens with bytes
    ),
]

Truthy = Annotated[
    Any,
    libcst.matchers.OneOf(
        libcst.matchers.Name(value="False"),
        libcst.matchers.Integer(value=_match_if_distinct("0")),
        libcst.matchers.Float(value=_match_if_distinct("0.0")),
        libcst.matchers.SimpleString(value=_match_if_distinct('""')),
    ),
]

# TODO: add literals for set, dict, tuple and list
# They will require to be itself Literal

# TODO: add HasCodeMatcher and DoesNotHaveCodeMatcher
